#!/usr/bin/env python2

# This is a simple script that can be used to visualize the results of a
# simulation and add more complex behaviour than is possible with a simple
# input data file.

# It is dependent on pygame for drawing output to screen. If this is not
# desirable, don't use the visualize module and resort to the outcommented
# calls to the fullsim module instead.

import apf
from apf import custom, fullsim, hooks
from apf.sims import visualize, simulation, groups, generate, optimizable
from apf.lib import ccalcs
import sys
import inspyred
import math
import random

class LineSimulation(simulation.Simulation):
    def done(self):
        lines = self.groupmanager.groups["line"]
        blue = self.groupmanager.groups["blue"]
        if not blue.objects:
            self.score += 1e10
        return (not blue.objects) or all(line.actuators[0].done for line in lines.objects)

class LineMover(object):

    def __init__(self, after, destination, speed):

        self.after = after
        self.destination = destination
        self.time = 0
        self.speed = speed
        self.done = False
        self.rotation = 2 * math.pi
        
    def __call__(self, obj, dt):

        self.time += dt
        if self.rotation <= 0:
            self.done = True
        if (obj.refpoint() - self.destination).length() < 1:
            angle = 0.0375 * dt
            obj.rotate(angle)
            self.rotation -= angle
        if self.after <= self.time:
            obj.data.velocity = self.velocity(obj, dt)
           
    def velocity(self, obj, dt):

        direction = self.destination - obj.refpoint()
        return direction.iscale_to(self.speed)

def make_simulation():

    simulation = custom.prepare_simulation(cls=LineSimulation)
    blue = simulation.groupmanager.groups["blue"]
    lines = simulation.groupmanager.groups["line"]

    orig = len(blue.objects)
    
    for unit in blue.objects:
        unit.add_hook(hooks.kill_close_units(2.5))
        unit.data.meta["radius"] = 1.25

    actuator = LineMover(25, ccalcs.Vector2(25, 15), 0.3)
    lines.objects[0].add_actuator(actuator)

    def distances():
        return sum([lines.objects[0].distance(obj.start).vector.length() for obj in blue.objects])

    def score(sim):
        if len(blue.objects) < 2: # Everyone is dead
            return 1000 # And that's really bad.
        # penalize average distance and units that have died
        timestep = distances() / len(blue.objects)
        return timestep

    def finalize(score, sim):
        return score / (sim.time + 1) + 30 * (orig - len(blue.objects))
        
    sc = optimizable.ScoreCounter(score, finalize)
    simulation.add_hook(sc.count)
    
    return simulation

RESOLUTION = (1200, 800) # The resolution this should be drawn to screen with
SIMULATION_SIZE = (120, 120) # The extend of the simulated world that will be drawn

def sim_printing_hook(candidate, score, sim):
    rows = ["$c_2$", "$\\sigma_2$", "$c_1$", "$\\sigma_1$", "Survived", "Performance"]
    outlist = candidate + [len(sim.groupmanager.groups["blue"].objects)] + [score]
    print " & ".join(rows) + "\\\\"
    print " & ".join(map(lambda x: "$%g$" % x, outlist)) + "\\\\" 
   
def main(argv):
    simulation = make_simulation()
    params = optimizable.InspyredInterface(make_simulation)
    params.add_train_force_law([(-2, -0.01)], [(-2, -0.1)], "line", "external")
    params.add_train_force_law([(0.01, 14)], [(0, 3)], "blue", "internal")
    if "visualize" in argv:
        visualization = visualize.PygameSimulation(simulation, RESOLUTION, SIMULATION_SIZE)
        visualization.framerate = 100
        visualization.run()
    elif "train" in argv:
        params.add_param("maximize", False)
        params.add_param("terminator", inspyred.ec.terminators.evaluation_termination)
        params.add_param("max_evaluations", 240)
        params.add_param("gaussian_stdev", 0.3)
        obs = [inspyred.ec.observers.stats_observer, inspyred.ec.observers.best_observer, inspyred.ec.observers.file_observer]
        params.add_param("observer", obs)
        classes = [inspyred.ec.ES, inspyred.ec.SA, inspyred.swarm.PSO, inspyred.ec.GA]
        pop_sizes = [40, 1, 12, 40]
        params_list = [params.copy() for _ in range(len(pop_sizes))]
        for index, pop_size in enumerate(pop_sizes):
            params_list[index].add_param("pop_size", pop_size)
        params_list[1].add_param("mutation_rate", 0.25) # High mutation rate for SA
        # Blend crossover for GA
        params_list[3].add_param("variator", [inspyred.ec.variators.blend_crossover])
        names, results = [], []
        for random_seed in range(10):
            _names, _results = optimizable.best_from(classes, params_list, random_seed)
            names.extend(_names)
            results.extend(_results)
        for name, result in zip(names, results):
            print name, 'found parameters: ', result[0], "which scored", result[1]
    elif len(argv) > 0 and argv[0] == "score":
        genome = map(float, argv[1:5])
        score = params.evaluate_sim(genome, {"sim_printing_hook": sim_printing_hook})
    elif len(argv) > 0 and argv[0] == "paint":
        genome = map(float, argv[1:5])
        sim = make_simulation()
        gm = sim.groupmanager.groups
        attraction = gm["line"].external.terms[0]
        attraction.c, attraction.sigma = genome[0], genome[1]
        repulsion = gm["blue"].internal.terms[0]
        repulsion.c, repulsion.sigma = genome[2], genome[3]
        sim.disable_dump = False
        sim.run()
        fullsim.full_sim_plot()
        fullsim.full_sim_movie()
    else:
        simulation.run()
        print "Simulation achieved score of", simulation.score
        fullsim.full_sim_plot()
        fullsim.full_sim_movie()

if __name__ == "__main__":
    main(sys.argv[1:])
