#!/usr/bin/env python2

# This is a simple script that can be used to visualize the results of a
# simulation and add more complex behaviour than is possible with a simple
# input data file.

# It is dependent on pygame for drawing output to screen. If this is not
# desirable, don't use the visualize module and resort to the outcommented
# calls to the fullsim module instead.

from apf import custom, hooks
from apf.sims import visualize, simulation, groups, optimizable
from apf.lib import ccalcs
import sys
import inspyred
import random
import math

added_units = [0]
add_unit_interval = 45

def make_simulation():
    simulation = custom.prepare_simulation()
# Add objects or modify them here.

    gm = simulation.groupmanager
    blue = gm.groups["blue"]
    red = groups.SimObjectGroup(ccalcs.makeSquare(-60, -60, 120), tag="red")
    for obj in red.objects:
        obj.data.meta['plt'] = 'r-'

    yellow = groups.SimObjectGroup([], tag="yellow")
    yellow.external = ccalcs.ForceLaw([ccalcs.MindistancePotential(-0.0375, -0.05, 9)])
    gm.groups["yellow"] = yellow
    gm.relations.add(("red", "blue"))
    
    def bounding_box_corners():
        side_lengths = ccalcs.bounding_box_side_lengths(blue.objects, 2)
        halflength = max(side_lengths.x, side_lengths.y) / 2
        center = ccalcs.center_of_mass(blue.objects)
        diag = ccalcs.Vector2(halflength, halflength)
        lower_left, upper_right = center - diag, center + diag
        upper_left = center + ccalcs.Vector2(-halflength, halflength)
        lower_right = center + ccalcs.Vector2(halflength, - halflength)
        return lower_left, lower_right, upper_left, upper_right

    def add_hooks(obj):
        obj.add_hook(hooks.kill_close_units(3))
        obj.add_actuator(ccalcs.accelerating_actuator)
        obj.add_actuator(hooks.friction_hook(0.75))
        obj.data.meta["radius"] = 1.5

    simulation.max_length = len(blue.objects)
    def check_finished_hook(sim):
        if len(blue.objects) > sim.max_length:
            sim.max_length = len(blue.objects)
            print "Increased sim.max_length, is now", sim.max_length            
        elif len(blue.objects) < sim.max_length:
            sim.finish()

    simulation.add_hook(check_finished_hook)
    
    def add_units():
        positions = [(40, 40), (-40, 40), (-40, -40), (40, -40)]
        for position in positions:
            unit = ccalcs.SimObject(ccalcs.Vector2(*position))
            add_hooks(unit)
            blue.add(unit)

    def add_units_hook(sim):
        if sim.dt - added_units[-1] > add_unit_interval:
            added_units.append(sim.dt)
            add_units()

    simulation.add_hook(add_units_hook)
    
    for point in bounding_box_corners():
        newobj = ccalcs.SimObject(point)
        newobj.data.meta["plt"] = "yo"
        newobj.data.meta["radius"] = 1.5
        yellow.add(newobj)
        
    def mindistance_hook(yellow):
        try:
            sides = ccalcs.bounding_box_side_lengths(blue.objects, 2)
        except IndexError:
            return # Oops, out of units
        sidel = max(sides.x, sides.y)
        yellow.external.terms[0].mindistance = sidel / 2

    yellow.add_hook(mindistance_hook)
    yellow.add_hook(hooks.square_moving_hook(bounding_box_corners, setend=False))

    red_move_hook = hooks.square_moving_hook(bounding_box_corners)
    red_move_hook(red)
    red.add_hook(red_move_hook)
    gm.groups["red"] = red
        
    for obj in blue.objects:
        add_hooks(obj)

    orig = len(blue.objects)
    
    def count_score(sim):
        sides = ccalcs.bounding_box_side_lengths(blue.objects, 2)
        if sides.product() > 10000:
            sim._done = True
        return (max(sides.x, sides.y) ** 2 + 225 * (sim.max_length - len(blue.objects))) / (len(blue.objects) * math.pi * 4)
        
    def final_score(score, sim):
        if len(blue.objects) == 0:
            sim.finish()
            return 10000
        return (score / (sim.time + 1))# / ((sim.max_length - orig) / 4 + 1)
       
    simulation.add_hook(optimizable.ScoreCounter(count_score, final_score).count)
    
    return simulation

def visualize_(sim=None):
    if sim is None:
        sim = make_simulation()
    RESOLUTION = (1200, 800) # The resolution this should be drawn to screen with
    SIMULATION_SIZE = (120, 120) # The extend of the simulated world that will be drawn
    visualization = visualize.PygameSimulation(sim, RESOLUTION, SIMULATION_SIZE)
    visualization.framerate = 100
    visualization.run()

force_law_getters = [optimizable.force_law_getter('blue', 'internal'),
                     optimizable.force_law_getter('yellow', 'external')]

force_law_table_header = (' & '.join('$%s$' % i for i in ['c_1', '\\sigma_1', 'c_2', '\\sigma_2',
                                                          'c_3', '\\sigma_3']) +
                          ' & Survivors & Performance \\\\')
    
def evaluate(candidate, args=None):
    simulation = make_simulation()
    optimizable.apply_genome(candidate, simulation, force_law_getters)
    simulation.disable_dump = True
    try:
        simulation.run()
    except ZeroDivisionError:
        return 10000
    except OverflowError:
        return 10000
    except IndexError:
        return 10000
    return simulation.score

      
def evaluate_and_dump_table(params):
    sim = make_simulation()
    optimizable.apply_genome(params, sim, force_law_getters)
    simulation.disable_dump = True
    sim.run()
    performance = sim.score
    survivors = sim.max_length
    print force_law_table_header
    print ' & '.join(['$%g$' % i for i in params + [survivors, performance]]) + '\\\\'

def inspyred_():
    params = optimizable.InspyredInterface(make_simulation)
    params.add_train_force_law([(-3, -0.001), (0, 30)], [(-3, 3), (-2, 5)], "blue", "internal")
    params.add_train_force_law([(-2, -0.001)], [(-2, -0.001)], "yellow", "external")
    params.add_param("maximize", False)
    params.add_param("max_evaluations", 800)
    params.add_param("gaussian_stdev", 0.1)
    obs = [inspyred.ec.observers.file_observer, inspyred.ec.observers.best_observer]
    params.add_param("observer", obs)
    params.add_param("terminator", inspyred.ec.terminators.evaluation_termination)    
    return params

def benchmark_optimize(seeds, opt_type):
    for seed in seeds:
        rand = random.Random()
        rand.seed(seed)
        params = inspyred_()
        if opt_type == "sa":
            ec = inspyred.ec.SA(rand)
            pop_size = 1
            params.add_param("mutation_rate", 0.2)
        elif opt_type == "es":
            ec = inspyred.ec.ES(rand)
            pop_size = 20
        elif opt_type == "pso":
            ec = inspyred.swarm.PSO(rand)
            pop_size = 12
        elif opt_type == "ga":
            ec = inspyred.ec.GA(rand)
            pop_size = 100
            params.add_param('variator', [inspyred.ec.variators.blend_crossover])
        params.add_param('pop_size', pop_size)
        params.add_param('statistics_file', open("%s%s.csv" % (opt_type.upper(), seed), "w"))
        params.add_param('individuals_file', open("%s%s-individuals.csv" % (opt_type.upper(), seed), "w"))
        kwargs = params.make_kwargs()
        for key in ["terminator", "variator", "observer", "archiver"]:
            if key in kwargs:
                setattr(ec, key, kwargs[key])
        pop = params.apply_and_run(ec, evaluate, startmp=(seed == seeds[0]), stopmp=(seeds == seeds[-1]))
        pop.sort(reverse=True)
        print pop[0].candidate, pop[0].fitness

best_params = {
    "pso": [-1.1293081252129489, 0.17303892971682913, 17.313650341904207, 1.3185743761724364, -0.5322908875729837, -0.37882621961362845],
    "ga": [-0.9364592500711375, 0.28536699463530557, 14.12297971482358, 1.4454022941123612, -0.37233959427987007, -0.2941715310412301],
    "es": [-0.001, -1.3858209081487192, 9.004028586403413, 2.343957281283811, -0.001, -0.001],
    "sa": [-0.0695459613950775, -0.36928239038898986, 11.775179223920906, 1.831888887515647, -0.37730699972518567, -0.3816491275825995],
    "human": [-1, 1, 9, 2, -0.0375, -0.05],
    "ga2": [-0.924306617433727, 0.157413463544687, 8.9072468466753, 1.239875921576678, -0.31394933353280385, -0.44386323018058543],
    "ga3": [-1.2455439369463406, -0.13007798971428486, 19.75783801499139, 0.7622358237397214, -0.7840583492395883, -0.5765630744103156],
    "ga4": [-1.4496458614371923, -0.09653553297380016, 17.69140125134309, 0.8083442942474866, -0.8602976522213654, -0.8910725135938217]
}

def score_them():
    for params in best_params:
        print params.upper()
        evaluate_and_dump_table(best_params[params])
        
def main(argv):
    if len(argv) == 0:
        from apf import fullsim
        make_simulation().run()
        fullsim.full_sim_plot()
        fullsim.full_sim_movie()
    elif argv[0] == "run":
        params = best_params[argv[1]]
        sim = make_simulation()
        optimizable.apply_genome(best_params[argv[1]], sim, force_law_getters)
        sim.run()
    elif argv[0] == "visualize":
        if len(argv) == 1:
            visualize_()
        else:
            sim = make_simulation()
            optimizable.apply_genome(best_params[argv[1]], sim, force_law_getters)
            visualize_(sim)
    elif argv[0] == "score":
        try:
            params = [float(arg) for arg in argv[1:]]
            evaluate_and_dump_table(params)
        except ValueError:
            params = best_params[argv[1]]
            evaluate_and_dump_table(params)
    elif argv[0] == "score_them":
        score_them()
    elif argv[0] == "paint":
        simulation = make_simulation()
        optimizable.apply_genome(best_params[argv[1]], simulation, force_law_getters)
        simulation.run()
        print ccalcs.bounding_box_side_lengths(simulation.groupmanager.groups["blue"].objects, 2)
        from apf import fullsim
        fullsim.full_sim_plot()
        fullsim.full_sim_movie()
    else:
        benchmark_optimize([1, 2, 7, 9], argv[0])
        
if __name__ == "__main__":
    main(sys.argv[1:])
