#!/usr/bin/env python2

# This is a simple script that can be used to visualize the results of a
# simulation and add more complex behaviour than is possible with a simple
# input data file.

# It is dependent on pygame for drawing output to screen. If this is not
# desirable, don't use the visualize module and resort to the outcommented
# calls to the fullsim module instead.

from apf import custom, fullsim, hooks
from apf.sims import visualize, simulation, optimizable
from apf.lib import ccalcs
import sys
import math
import random
import itertools

speeds = itertools.cycle([0.3, 0.35, 0.4])

def pick_speed(unit):
    unit.data.meta["max_speed"] = speeds.next()

class LineMover(object):

    def __init__(self):

        self.speed = 0
        self.destination = ccalcs.Vector2(0, 0)
        self.arrived = False
        self.callback = None
        
    def __call__(self, line):

        if self.speed and not self.arrived:
            line.data.velocity = self.velocity(line)

    def velocity(self, obj):

        direction = self.destination - obj.refpoint()
        if direction.length() < self.speed:
            self.arrived = True
            if self.callback is not None:
                self.callback(self, obj)
        try:
            return direction.iscale_to(self.speed)
        except ZeroDivisionError:
            return ccalcs.Vector2(0, 0)

    def set_destination(self, dest):

        self.arrived = False
        self.destination = dest

    def on_arrival(self, func):

        self.callback = func

genes = {
    "pso": [-0.992032, -0.334872, 14, 2.87928],
    "sa": [-0.409996, -0.333396, 10.2848, 3],
    "ga": [-0.8736, -0.960802, 10.48985, 2.33163],
    "es": [-0.666186, -0.480902, 9.60076, 2.62909],
    "human": [-0.5, -0.5, 7, 2.3]
}

fl_getters = [optimizable.force_law_getter("red", "external"),
              optimizable.force_law_getter("blue", "internal")]

def make_simulation():
    simulation = custom.prepare_simulation()

    gm = simulation.groupmanager

    blue = gm.groups["blue"]
    red = gm.groups["red"]
    green = gm.groups["green"]

    line = red.objects[0]
    
    mover = LineMover()
    mover.speed = 0.25

    def set_mover(sim):
        line.add_hook(mover)
        mover.set_destination(ccalcs.Vector2(20, 0))
        def finish(*args):
            simulation.finish()
        def activate_new(sim):
            mover.set_destination(ccalcs.Vector2(20, 40))
            mover.on_arrival(finish)
        def start(*args):
            sim.add_hook(hooks.delayed_action(simulation.dt + 30, activate_new))
        mover.on_arrival(start)

    simulation.add_hook(hooks.delayed_action(50, set_mover))
    
    def distances():
        return sum([line.distance(obj.refpoint()).vector.length() for obj in blue.objects])

    def score_f(sim):
        if len(blue.objects) < 2:
            return 1000
        return distances() / len(blue.objects)

    def finalize(score, sim):
        return score / (sim.time + 1) + 30 * (12 - len(blue.objects))

    simulation.add_hook(optimizable.ScoreCounter(score_f, finalize).count)
    
    def is_close_to_green(obj):
        return len(green.objects_near(obj.refpoint(), obj.data.meta['radius'] + 6)) > 0

    def obj_on_green(obj):
        if not "old_speed" in obj.data.meta:
            obj.data.meta["old_speed"] = obj.data.meta["max_speed"]
            obj.data.meta["max_speed"] = 0.2

    def obj_off_green(obj):
        if "old_speed" in obj.data.meta:
            obj.data.meta["max_speed"] = obj.data.meta["old_speed"]
            del obj.data.meta["old_speed"]

    def obj_on_green_hook(obj):
        if is_close_to_green(obj):
            obj_on_green(obj)
        else:
            obj_off_green(obj)

    def add_hooks(obj):
        obj.data.meta["radius"] = 1.25
        obj.add_hook(hooks.kill_close_units(2.5))
        obj.add_hook(obj_on_green_hook)
        pick_speed(obj)
        
    for obj in blue.objects:
        add_hooks(obj)

    def add_two_units_hook(sim):
        one, two = ccalcs.SimObject(ccalcs.Vector2(-25, -40)), ccalcs.SimObject(ccalcs.Vector2(-30, -40))
        add_hooks(one)
        add_hooks(two)
        one.data.meta["plt"] = "bo"
        two.data.meta["plt"] = "bo"
        blue.add(one)
        blue.add(two)

    simulation.add_hook(hooks.delayed_action(150, add_two_units_hook))
        
    return simulation

def viz(framerate=100, sim=None):
    if sim is None:
        sim = make_simulation()
    RESOLUTION = (1200, 800) # The resolution this should be drawn to screen with
    SIMULATION_SIZE = (120, 120) # The extend of the simulated world that will be drawn
    visualization = visualize.PygameSimulation(sim, RESOLUTION, SIMULATION_SIZE)
    visualization.framerate = framerate
    visualization.run()

def main(argv):
    if not argv:
        sim = make_simulation()
        viz(sim=sim)
    elif argv[0] == "run":
        sim = make_simulation()
        sim.run()
        print sim.score
    elif argv[0] == 'animate':
        fullsim.full_sim_plot()
        fullsim.full_sim_movie()
    elif argv[0] == 'paint':
        sim = make_simulation()
        optimizable.apply_genome(genes[argv[1]], sim, fl_getters)
        sim.run()
    elif argv[0] == 'score_them':
        print "Technique & $c_{2}$ & $\sigma_{2}$ & $c_{1}$ & $\sigma_{2}$ & $P(S)$ \\\\"
        for key in genes:
            sim = make_simulation()
            sim.disable_dump = True
            optimizable.apply_genome(genes[key], sim, fl_getters)
            sim.run()
            out = (key + " & " + " & ".join("$%g$" % num for num in genes[key] + [sim.score]) + "\\\\")
            print out
        
if __name__ == "__main__":

    main(sys.argv[1:])
