#!/usr/bin/env python2

# This is a simple script that can be used to visualize the results of a
# simulation and add more complex behaviour than is possible with a simple
# input data file.

# It is dependent on pygame for drawing output to screen. If this is not
# desirable, don't use the visualize module and resort to the outcommented
# calls to the fullsim module instead.

from apf import custom, fullsim, geometry, hooks
from apf.sims import visualize, simulation, groups, optimizable
from apf.lib import ccalcs
import sys
import itertools

speeds = itertools.cycle([0.45, 0.4, 0.5])

TRIANGLE_SPEED = 0.15

def pick_speed(unit):
    unit.data.meta["max_speed"] = speeds.next()

def make_simulation():
    simulation = custom.prepare_simulation()
    gm = simulation.groupmanager

    blue = gm.groups["blue"]

    obstacles = [(0, 0), (10, -5), (20, 10), (-5, -10), (-10, 15), (10, 20)]
    green = groups.SimObjectGroup([], tag="green")
    for point in obstacles:
        obj = ccalcs.SimObject(ccalcs.Vector2(*point))
        obj.data.meta["plt"] = "go"
        obj.data.meta["radius"] = 4
        green.add(obj)

    gm.groups["green"] = green
    
    def close_to_green(obj):
        return len(green.objects_near(obj.refpoint(), obj.data.meta["radius"] + 4)) > 0

    def obj_on_green(obj):
        if not "old_speed" in obj.data.meta:
            obj.data.meta["old_speed"] = obj.data.meta["max_speed"]
            obj.data.meta["max_speed"] = 0.1

    def obj_off_green(obj):
        if "old_speed" in obj.data.meta:
            obj.data.meta["max_speed"] = obj.data.meta["old_speed"]
            del obj.data.meta["old_speed"]

    def obj_on_green_hook(obj):
        if close_to_green(obj):
            obj_on_green(obj)
        else:
            obj_off_green(obj)

    def add_hooks(unit):
        pick_speed(unit)
        unit.add_actuator(hooks.friction_hook(0.5))
        unit.add_actuator(ccalcs.accelerating_actuator)
        unit.add_hook(hooks.kill_close_units(4))
        unit.add_hook(obj_on_green_hook)
        unit.data.meta["radius"] = 2
        unit.data.meta["plt"] = "bo"
        
    for unit in blue.objects:
        add_hooks(unit)
        
    corners = [ccalcs.Vector2(x, y) for x, y in [(-50, -40), (-50, 40), (-10, 0)]]
    bottom, top, right = corners

    triangle = geometry.Triangle(*corners)

    def triangle_move_hook(sim):
        if 90 < sim.dt < 260:
            triangle.move(ccalcs.Vector2(1, 0).normalize().scale(TRIANGLE_SPEED * sim.params.timestep))
        elif 260 < sim.dt < 440:
            triangle.move(ccalcs.Vector2(1, 1).normalize().scale(TRIANGLE_SPEED * sim.params.timestep))
        elif 440 < sim.dt < 460:
            triangle.move(ccalcs.Vector2(1, -1).normalize().scale(TRIANGLE_SPEED * sim.params.timestep))
        elif 600 < sim.dt < 780:
            triangle.move(ccalcs.Vector2(-1, 0).normalize().scale(TRIANGLE_SPEED * sim.params.timestep))
        elif sim.dt > 800:
            sim.finish()
            
    simulation.add_hook(triangle_move_hook)

    max_units = [float(len(blue.objects))]
    
    def count_score(sim):
        amount = len([obj for obj in blue.objects if triangle.contains(obj.refpoint())])
        return float(amount) / max_units[0]

    def final_score(score, sim):
        return score / (sim.time + 1)

    simulation.add_hook(optimizable.ScoreCounter(count_score, final_score).count)
    
    red = groups.SimObjectGroup([ccalcs.SimObject(bottom, top), ccalcs.SimObject(bottom, right), ccalcs.SimObject(top, right)], tag="red")

    gm.groups["red"] = red

    for obj in red.objects:
        obj.data.meta["plt"] = "r-"

    com = triangle.center_of_mass()
    yellow = groups.SimObjectGroup(map(lambda p: ccalcs.SimObject(p, com, True), corners), tag="yellow")

    bind_endpoint = hooks.bind_endpoint_hook(triangle.center_of_mass)
    
    for obj in yellow.objects:
        obj.data.meta["plt"] = "y-"
        bind_endpoint(obj)
        obj.add_hook(bind_endpoint)
        
    gm.groups["yellow"] = yellow
    yellow.external.terms = [ccalcs.PotentialTerm(-0.5, -1)]

    gm.relations.add(("yellow", "blue"))

    def add_three_units_hook(sim):
        positions = [(-40, 25), (-50, 20), (-35, 30)]
        units = [ccalcs.SimObject(ccalcs.Vector2(*pos)) for pos in positions]
        max_units[0] += 3
        for unit in units:
            add_hooks(unit)
            blue.add(unit)

    simulation.add_hook(hooks.delayed_action(200, add_three_units_hook))
    
    return simulation

def viz(sim=None, framerate=100):
    if sim is None:
        sim = make_simulation()
    RESOLUTION = (1200, 1024) # The resolution this should be drawn to screen with
    SIMULATION_SIZE = (160, 160) # The extend of the simulated world that will be drawn
    visualization = visualize.PygameSimulation(sim, RESOLUTION, SIMULATION_SIZE)
    visualization.framerate = framerate
    visualization.run()
    print "Scored:", sim.score

fl_getters = [optimizable.force_law_getter("blue", "internal"),
              optimizable.force_law_getter("yellow", "external")]

genes = {
    "human": [9, 1, -0.5, -1],
    "sa": [4.57825, 1.22634, -0.183973, -0.929281],
    "ga": [12.7609, 1.08366, -0.696471, -0.914919],
    "pso": [12.8845, 0.856295, -0.892512, -1.02084],
    "es": [14.3339, 1.21731, -0.520555, -0.907055],
    "custom": [20, 1.25, -0.5, -0.5]
}

def get_sim_for_gene(name):
    sim = make_simulation()
    optimizable.apply_genome(genes[name], sim, fl_getters)
    sim.disable_dump = True
    sim.run()
    return sim

def latexify_num(n):
    if isinstance(n, int):
        return "$%d$" % n
    else:
        return "$%g$" % n
        
def main(argv):
    if not argv:
        viz()
    elif argv[0] == "run":
        sim = make_simulation()
        if len(argv) > 1:
            optimizable.apply_genome(genes[argv[1]], sim, fl_getters)
        sim.run()
        print "Scored:", sim.score
    elif argv[0] == "animate":
        fullsim.full_sim_plot()
        fullsim.full_sim_movie()
    elif argv[0] == "test":
        sim = make_simulation()
        optimizable.apply_genome(genes[argv[1]], sim, fl_getters)
        if not "dump" in argv:
            sim.disable_dump = True
        sim.run()
        print sim.time, sim.dt
        print argv[1], "=>", sim.score, len(sim.groupmanager.groups["blue"].objects)
    elif argv[0] == "score_them":
        print "Technique & $c_{1}$ & $\\sigma_{1}$ & $c_{2}$ & $\\sigma_{2}$ & P(S) & Survivors\\\\"
        for gene in genes:
            sim = get_sim_for_gene(gene)
            out = gene.upper() + ' & '
            numbers = genes[gene]
            numbers.append(sim.score)
            numbers.append(len(sim.groupmanager.groups["blue"].objects))
            print out + " & ".join(map(latexify_num, numbers)) + "\\\\"
        
if __name__ == "__main__":
    main(sys.argv[1:])