#!/usr/bin/env python2

# This is a simple script that can be used to visualize the results of a
# simulation and add more complex behaviour than is possible with a simple
# input data file.

# It is dependent on pygame for drawing output to screen. If this is not
# desirable, don't use the visualize module and resort to the outcommented
# calls to the fullsim module instead.

from apf import custom, fullsim, hooks
from apf.sims import visualize, simulation, groups, optimizable
from apf.lib import ccalcs
from apf import geometry
import sys
import math
import inspyred
import random

def make_simulation():
    simulation = custom.prepare_simulation()
    gm = simulation.groupmanager
    blue = gm.groups["blue"]

    orig = float(len(blue.objects))
    
    sides = geometry.bounding_box_side_lengths(blue.objects, 2)
    side = max(sides.x, sides.y)

    base_left, base_right = ccalcs.Vector2(-1 * side, 0), ccalcs.Vector2(1 * side, 0)
    top = ccalcs.Vector2(0, side)

    red = groups.SimObjectGroup([], tag="red")
    red.add(ccalcs.SimObject(base_left, base_right))
    red.add(ccalcs.SimObject(base_left, top))
    red.add(ccalcs.SimObject(base_right, top))
    
    for obj in red.objects:
        obj.data.meta["plt"] = "r-"
    gm.groups["red"] = red

    corner_pull = ccalcs.ForceLaw([ccalcs.PotentialTerm(-0.5, -1)])
    
    yellow = groups.SimObjectGroup([], tag="yellow", external=corner_pull)
    com = ccalcs.center_of_mass([base_left, base_right, top])
    yellow.add(ccalcs.SimObject(base_left, com, True))
    yellow.add(ccalcs.SimObject(base_right, com, True))
    yellow.add(ccalcs.SimObject(top, com, True))

    gm.relations.add(("yellow", "blue"))
    
    for obj in yellow.objects:
        obj.data.meta["plt"] = "y-"
    gm.groups["yellow"] = yellow
    
    triangle = geometry.Triangle(base_left, base_right, top)

    def move_hook(sim):
        triangle.move(ccalcs.Vector2(0, -0.005))
    
    for obj in blue.objects:
        obj.add_hook(hooks.kill_close_units(4))
        obj.add_actuator(ccalcs.accelerating_actuator)
        obj.add_actuator(hooks.friction_hook(0.75))
        obj.data.meta["radius"] = 2
       
    def count_score(sim):
        amount = len([obj for obj in blue.objects if triangle.contains(obj.refpoint())])
        if amount == 0:
            sim.finish()
        return amount / orig

    def final_score(score, sim):
        if len(blue.objects) == 0:
            sim.finish()
            return 0
        return score / (sim.time + 1)

    simulation.add_hook(optimizable.ScoreCounter(count_score, final_score).count)

    return simulation
    
    
# Add objects or modify them here.

RESOLUTION = (1200, 1024) # The resolution this should be drawn to screen with
SIMULATION_SIZE = (120, 120) # The extend of the simulated world that will be drawn

def visualize_(sim=None, framerate=100):
    if sim is None:
        sim = make_simulation()
    visualization = visualize.PygameSimulation(sim, RESOLUTION, SIMULATION_SIZE)
    visualization.framerate = framerate
    visualization.run()

force_law_getters = [optimizable.force_law_getter('blue', 'internal'),
                     optimizable.force_law_getter('yellow', 'external')]
table_header = (' & '.join('$%s$' % c for c in ['c_{1}', '\\sigma_{1}', 'c_{2}', '\\sigma_{2}']) +
                ' & Survivors & $P(S)$ \\\\')

def evaluate(candidate, args=None):
    simulation = make_simulation()
    optimizable.apply_genome(candidate, simulation, force_law_getters)
    simulation.disable_dump = True
    try:
        simulation.run()
    except ZeroDivisionError:
        return 0
    except OverflowError:
        return 0
    except IndexError:
        return 0
    return simulation.score

def evaluate_and_dump(params):
    sim = make_simulation()
    optimizable.apply_genome(params, sim, force_law_getters)
    sim.disable_dump = True
    sim.run()
    print table_header
    s, p = len(sim.groupmanager.groups["blue"].objects), sim.score
    print ' & '.join(['$%g$' % param for param in (params + [s, p])]) + '\\\\'

def parse_and_evaluate(param_vector):
    params = map(float, param_vector)
    evaluate_and_dump(params)

def inspyred_():
    params = optimizable.InspyredInterface(make_simulation)
    params.add_train_force_law([(0, 20)], [(0, 2)], "blue", "internal")
    params.add_train_force_law([(-2, 0)], [(-2, 0)], "yellow", "external")
    params.add_param("maximize", True)
    params.add_param("max_evaluations", 300)
    params.add_param("gaussian_stdev", 0.1)
    obs = [inspyred.ec.observers.file_observer]
    params.add_param("observer", obs)
    params.add_param("terminator", inspyred.ec.terminators.evaluation_termination)
    params.add_param("archiver", inspyred.ec.archivers.best_archiver)
    return params

def optimize(seeds, opt_type):
    for seed in seeds:
        rand = random.Random()
        rand.seed(seed)
        params = inspyred_()
        if opt_type == "sa":
            ec = inspyred.ec.SA(rand)
            pop_size = 1
            params.add_param("mutation_rate", 0.25)
        elif opt_type == "es":
            ec = inspyred.ec.ES(rand)
            pop_size = 40
        elif opt_type == "pso":
            ec = inspyred.swarm.PSO(rand)
            pop_size = 12
        elif opt_type == "ga":
            ec = inspyred.ec.GA(rand)
            pop_size = 40
            params.add_param("variator", [inspyred.ec.variators.blend_crossover])
        params.add_param("pop_size", pop_size)
        params.add_param("statistics_file", open("%s%s.csv" % (opt_type.upper(), seed), "w"))
        params.add_param("individuals_file", open("%-sindividuals-%s.csv" % (opt_type.upper(), seed), "w"))
        kwargs = params.make_kwargs()
        for key in ["terminator", "variator", "observer", "archiver"]:
            if key in kwargs:
                setattr(ec, key, kwargs[key])
        pop = params.apply_and_run(ec, evaluate, startmp=(seed == seeds[0]), stopmp=(seed == seeds[-1]))
        pop.sort()
        best = ec.archive[0]
        print best.candidate, best.fitness

best_params = {
    "pso": [12.88450302528144, 0.8562946312663792, -0.8925119504328345, -1.020836645894892],
    "es": [14.333895328234995, 1.2173052703702625, -0.5205552240246245, -0.9070549553230037],
    "sa": [4.578249123084022, 1.2263405306073483, -0.18397251650478263, -0.9292809965014612],
    "ga": [12.760851168779585, 1.0836592304089445, -0.6964705318803888, -0.9149188184602308]
}

def main(argv):
    if not argv:
        visualize_()
    elif argv[0] == 'run':
        make_simulation().run()
    elif argv[0] == 'animate':
        fullsim.full_sim_plot()
        fullsim.full_sim_movie()
    elif argv[0] == "score":
        parse_and_evaluate(argv[1:])
    elif argv[0] == "show":
        params = map(float, argv[1:])
        sim = make_simulation()
        optimizable.apply_genome(params, force_law_getters)
        visualize_(sim)
    elif argv[0] == "optimize":
        optimize(range(10), argv[1])
    elif argv[0] == "paint":
        params = best_params[argv[1]]
        sim = make_simulation()
        optimizable.apply_genome(params, sim, force_law_getters)
        sim.run()
        fullsim.full_sim_plot()
        fullsim.full_sim_movie()
       
if __name__ == "__main__":
    main(sys.argv[1:])
