"""
Some hooks for using with simulations and/or objects.
"""

def kill_close_units(distance):
    def kill_hook(simobj):
        near_units = simobj.group.objects_near(simobj.refpoint(), distance)
        for unit in near_units:
            if unit != simobj:
                simobj.group.kill_unit(unit)
        if len(near_units) > 1:
            simobj.group.kill_unit(simobj)
    return kill_hook

def square_moving_hook(corners_func, setend=True):
    def move_hook(group):
        try:
            lower_left, lower_right, upper_left, upper_right = corners_func()
        except IndexError: # Oops, out of units
            return 
        top, bottom, left, right = group.objects
        bottom.start = lower_left
        left.start = upper_left
        right.start = lower_right
        top.start = upper_right
        if setend:
            top.end = upper_left
            right.end = upper_right
            bottom.end = lower_right
            left.end = lower_left
    return move_hook

def bind_endpoint_hook(get_endpoint):
    def bind_endpoint(obj):
        obj.end = get_endpoint()
    return bind_endpoint
    
def friction_hook(factor):
    def slow_down(obj, dt):
        obj.data.velocity.iscale(factor)
    return slow_down

class ActivatedHook(object):

    def __init__(self, activate_cond, deactivate_cond, hook_func):

        self.activated = False
        self.deactivated = False
        self.start = activate_cond
        self.stop = deactivate_cond
        self.run = hook_func
        
    def __call__(self, *args, **kwargs):

        if self.should_be_active(*args, **kwargs):
            self.run(*args, **kwargs)
        self.maybe_shutdown(*args, **kwargs)

    def should_be_active(self, *args, **kwargs):
        if not self.activated:
            if self.start(*args, **kwargs):
                self.activated = True
        return self.activated and (not self.deactivated)

    def maybe_shutdown(self, *args, **kwargs):
        if self.stop(*args, **kwargs):
            self.deactivated = True

def is_near_check(point, obj):
    def near_check(sim):
        return point.distance_length(obj.refpoint()) < 1e-6
    return near_check

def move_hook(direction):
    def move(obj, dt):
        obj.data.velocity = direction
    return move

def delay_check(delay_time):
    def delay(sim):
        return sim.dt >= delay_time
    return delay

def delayed_action(delay, action):
    finished = [False]
    check = delay_check(delay)
    def run_hook(sim):
        if check(sim) and not finished[0]:
            action(sim)
            finished[0] = True
    return run_hook

def run_once(activation_cond, activation_func):
    return ActivatedHook(activation_cond, lambda _: True, activation_func)
    