cdef extern from "math.h":
    double sqrt(double)
    double atan2(double, double)
    double sin(double)
    double cos(double)
    double acos(double)
    
cdef class Vector2:
    """A 2-dimensional vector class for euclidian geometry. Methods that
    start with the letter i acts in-place and will modify the vector.

    Most methods here will return a Vector2. Overloaded arithmetic methods
    only accept parameters for wich isinstance(param, Vector2) is True."""
    
    cdef public double x
    cdef public double y

    def dict(self):
        """Serialize a Vector2 to a dict containing keys x, y."""
        return {"x": self.x, "y": self.y}
        
    def __init__(self, double x, double y):

        self.x = x
        self.y = y

    def __add__(Vector2 self, Vector2 other):
        return Vector2(self.x + other.x, self.y + other.y)

    def __iadd__(Vector2 self, Vector2 other):
        self.x += other.x
        self.y += other.y
        return self
        
    def __sub__(Vector2 self, Vector2 other):
        return Vector2(self.x - other.x, self.y - other.y)

    def __isub__(Vector2 self, Vector2 other):
        self.x -= other.x
        self.y -= other.y
        return self
        
    def __mul__(Vector2 self, Vector2 other):
        return Vector2(self.x * other.x, self.y * other.y)

    def __imul__(Vector2 self, Vector2 other):
        self.x *= other.x
        self.y *= other.y
        return self

    def __neg__(Vector2 self):
        return Vector2(-self.x, -self.y)

    def __richcmp__(Vector2 self, Vector2 other, int cmp):
        cdef double mSquared = self.length_squared()
        cdef double oSquared = other.length_squared()
        if cmp == 0:
            return mSquared < oSquared
        elif cmp == 2:
            return self.x == other.x and self.y == other.y
        elif cmp == 4:
            return mSquared > oSquared
        elif cmp == 1:
            return mSquared <= oSquared
        elif cmp == 3:
            return self.x != other.x or self.y != other.y
        elif cmp == 5:
            return mSquared >= oSquared
        
    cpdef double length_squared(Vector2 self):
        return self.x * self.x + self.y * self.y

    cpdef double length(Vector2 self):
        return sqrt(self.length_squared())

    cpdef Vector2 divide_by(Vector2 self, double number):
        return Vector2(self.x / number, self.y / number)

    cpdef Vector2 idivide_by(Vector2 self, double number):
        self.x /= number
        self.y /= number
        return self

    cpdef Vector2 normalize(Vector2 self):
        return self.divide_by(self.length())

    cpdef Vector2 inormalize(Vector2 self):
        return self.idivide_by(self.length())
           
    cpdef double angle(Vector2 self):
        """<=> atan2(y, x)"""
        return atan2(self.y, self.x)

    cpdef double sin(Vector2 self) except? 1.0:
        """<=> sin(x / y)"""
        return sin(self.x / self.y)

    cpdef double rsin(Vector2 self) except? 1.0:
        """<=> sin(y / x)"""
        return sin(self.y / self.x)

    cpdef double cos(Vector2 self) except? 1.0:
        """<=> cos(x / y)"""
        return cos(self.x / self.y)

    cpdef double rcos(Vector2 self) except? 1.0:
        """<=> cos(y / x)"""
        return cos(self.y / self.x)

    cpdef Vector2 irotate(Vector2 self, double angle):
        """Rotate self counterclockwise by angle. This rotates around the origin."""
        cdef double sina, cosa, x
        sina = sin(angle)
        cosa = cos(angle)
        x = self.x * cosa - self.y * sina
        self.y = self.x * sina + self.y * cosa
        self.x = x
        return self
    
    cpdef Vector2 rotate(Vector2 self, double angle):
        """Rotate self counterclockwise by angle. This rotates around the origin."""
        return Vector2(self.x, self.y).irotate(angle)

    cpdef Vector2 iperpendicular(Vector2 self):
        """Make self perpendicular to what it is now."""
        cdef double x
        x = self.x
        self.x = -self.y
        self.y = x
        return self

    cpdef Vector2 perpendicular(Vector2 self):
        """Find a vector that is perpendicular to self."""
        return Vector2(self.x, self.y).iperpendicular()

    cpdef double dot(Vector2 self, Vector2 other):
        """Dot product of two vectors."""
        return self.x * other.x + self.y * other.y

    cpdef Vector2 plus(Vector2 self, double arg):
        """<=> Vector2(x + arg, y + arg)"""
        return Vector2(self.x + arg, self.y + arg)

    cpdef Vector2 iplus(Vector2 self, double arg):
        self.x += arg
        self.y += arg
        return self

    cpdef Vector2 minus(Vector2 self, double arg):
        """<=> Vector2(x - arg, y - arg)"""
        return Vector2(self.x - arg, self.y - arg)

    cpdef Vector2 iminus(Vector2 self, double arg):
        self.x -= arg
        self.y -= arg

    cpdef Vector2 scale(Vector2 self, double factor):
        """<=> Vector2(x * factor, y * factor)"""
        return Vector2(self.x * factor, self.y * factor)

    cpdef Vector2 iscale(Vector2 self, double factor):
        self.x *= factor
        self.y *= factor
        return self
        
    def __str__(self):

        return "[%s, %s]" % (self.x, self.y)

    def __repr__(self):

        return "<Vector2: %s>" % (self,)

    cpdef Vector2 copy(self):
        return Vector2(self.x, self.y)

    cpdef Vector2 scale_down(Vector2 self, double maxlength):
        """Return a Vector2 that is scaled down to be no longer than param."""
        return self.copy().iscale_down(maxlength)

    cpdef Vector2 iscale_down(Vector2 self, double maxlength):
        """Scale down self so that self.length() <= maxlength.

        If self.length is already smaller than maxlength, take no action."""
        cdef double length = self.length()
        if length < maxlength:
            return self
        cdef double divide_by = length / maxlength
        return self.idivide_by(divide_by)

    cpdef Vector2 iscale_to(Vector2 self, double length):
        cdef double divide_by = self.length() / length
        return self.idivide_by(divide_by)

    cpdef Vector2 scale_to(Vector2 self, double length):
        return self.copy().iscale_to(length)
        
    cpdef double distance_length(Vector2 self, Vector2 other):
        cdef double x, y
        x = self.x - other.x
        y = self.y - other.y
        return sqrt(x * x + y * y)

    cpdef double product(Vector2 self):
        return self.x * self.y
       
ZERO = Vector2(0.0, 0.0)
ONE = Vector2(1.0, 1.0)
UNIT = ONE.normalize()
X = Vector2(1.0, 0)
Y = Vector2(0, 1.0)

cdef class MaybeVector2:
    """Used by methods and functions that may fail to calculate a
    Vector2 and need to signal this."""
    
    cdef public int isVector
    cdef public Vector2 vector
    
    def __init__(self, isVector, vector):
        self.isVector = isVector
        self.vector = vector

    def __str__(self):
        return "{%s, %s}" % (bool(self.isVector), self.vector)

    def __repr__(self):
        return "<MaybeVector2: %s>" % (self,)

    def dict(self):
        """Serialize a MaybeVector2 to a dict containing keys: isVector, vector."""
        return {"isVector": self.isVector, "vector": self.vector.dict()}
        
        
cdef class SimObjectData:
    """Container for the forces affecting a SimObject. Also allows
    for arbitrary metadata (name, colour, etc) to be attached to a SimObject."""
    
    cdef public Vector2 now, previously
    cdef public dict meta
    cdef public Vector2 velocity
        
    def __init__(self):
        self.now = Vector2(0, 0)
        self.previously = Vector2(0, 0)
        self.meta = {}
        self.velocity = Vector2(0, 0)
        
    cpdef tick(SimObjectData self):
        """Remember current force in self.previously and reset current force."""
        
        self.previously.x = self.now.x
        self.previously.y = self.now.y
        self.now -= self.now
        
    cpdef addForce(SimObjectData self, Vector2 force):

        self.now += force

    def tag(self, name):
        """Tag self with a name."""
        self.meta['name'] = name

    def __str__(self):

        return "Now: %s, Previously: %s, Velocity: %s\nMeta: %s" % (self.now, self.previously, self.velocity, self.meta)

    def __repr__(self):

        return "<SimObjectData: %s>" % (self,)

    def dict(self):
        """Serialize SimDataObject to a dict containing keys: now, previously, meta."""
        
        return {"now": self.now.dict(), "previously": self.previously.dict(), "meta": self.meta, "velocity": self.velocity.dict()}

    def __setitem__(self, key, value):

        self.meta[key] = value

    def __getitem__(self, key):

        return self.meta[key]

    def get(self, key, default=False):

        return self.meta.get(key, default)

cdef class SimObject:
    """SimObjects may be lines or pointmasses. They are used for calculating
    vectors between one another and for bookkeeping the results of force calculations
    in a SimObjectData object."""
    
    cdef public Vector2 start
    cdef public Vector2 end
    cdef public Vector2 slope
    cdef public double length
    cdef public SimObjectData data
    cdef public int round_edges
    cdef list hooks
    cdef public list actuators
    cdef public object group
    
    def __init__(self, start, end=None, round_edges=False):
        self.start = start
        self.end = end
        self._calcSlope()
        self._calcLength()
        self.data = SimObjectData()
        self.round_edges = round_edges
        self.hooks = []
        self.group = None
        self.actuators = []
        
    def add_hook(self, hook):
        self.hooks.append(hook)
        
    cdef double _calcSlope(SimObject self):
        if self.end is not None:
            self.slope = self.end - self.start

    cdef double _calcLength(SimObject self):
        if self.slope is not None:
            self.length = self.slope.length()
            
    cpdef int isLine(SimObject self):
        return self.end is not None

    cpdef MaybeVector2 centerOfMass(SimObject self, Vector2 reference_point):
        """Find the center of mass of this SimObject wrt. reference_point.

        If self is a line, find the point on the line that is closest to reference_point.
        If self does not have round_edges, this operation will not return a point if there
        is not a perpendicular line that intersects with both the line and the reference point."""
        
        if not self.isLine():
            return MaybeVector2(1, self.start)
        if self.slope is None:
            self._calcSlope()
            self._calcLength()
        return self._calcCenterOfMass(reference_point)

    cpdef move(SimObject self, Vector2 direction):
        """Move this object by adding direction to start and if necessary end point.""" 
        self.start += direction
        if self.isLine():
            self.end += direction

    cpdef add_actuator(self, actuator):
        """Add an actuator. An actuator is a callable of two parameters - the object it
        should actuate and the time that has been passed since the last frame of the
        simulation."""

        self.actuators.append(actuator)
        
    cpdef actuate(SimObject self, double dt):
        """Actuate this object. This runs all actuator hooks or adds force to velocity
        if no actuators have been set."""

        cdef int i
        acts = len(self.actuators)
        if acts == 0:
            self.data.velocity += self.data.now.scale(dt)
        else:
            for i in range(acts):
                self.actuators[i](self, dt)
                
    cdef MaybeVector2 _calcCenterOfMass(SimObject self, Vector2 reference_point):
        cdef double div, alpha, a, t, c
        cdef Vector2 startvec = reference_point - self.start
        c = startvec.length()
        div = c * self.length
        # We can't handle the case where reference_point coincides with self.start, as in, the
        # distance is identical to zero. Since the distance will be used for division, we signal
        # early on that it's probably a bad idea to be using this result.
        if div == 0: 
            return MaybeVector2(0, ZERO)
        # This uses the triangle identity: |a| |b| cos(alpha) = a.dot(b)
        # We use this to find the vector that is perpendicular to our line segment and crosses
        # through reference_point.
        # self.slope and (reference_point - self.start) are the two known sides which we use
        # the length of to calculate alpha. Use alpha to find the third side - the one
        # that intersects this line and reference_point
        alpha = acos(startvec.dot(self.slope) / div)
        a = c * cos(alpha)
        t = a / self.length
        if  0 <= t <= 1: # Find out if line is being intersected between start and end point.
            return MaybeVector2(1, self.start + self.slope.scale(t))
        else:
            return MaybeVector2(0, ONE) # Special case: if we have rounded_edges, we're not done yet

    cpdef MaybeVector2 distance(SimObject self, Vector2 point):
        """The shortes distance between self and point. If point intersects
        with self this may return a non-vector."""
        if self.round_edges:
            return self.distanceRoundEdges(point)
        else:
            return self.distanceFlat(point)
            
    cpdef MaybeVector2 distanceFlat(SimObject self, Vector2 point):
        """The distance from self to point, assuming that self does not have
        round_edges."""
        cdef MaybeVector2 me = self.centerOfMass(point)
        if me.isVector:
            return MaybeVector2(1, point - me.vector)
        else:
            return me

    cpdef MaybeVector2 distanceRoundEdges(SimObject self, Vector2 point):
        """The distance from self to point assuming that self has round_edges."""
        
        cdef MaybeVector2 me = self.centerOfMass(point)
        if me.vector is ONE:
            start, end = point - self.start, point - self.end
            if start.length() < end.length():
                return MaybeVector2(1, start)
            else:
                return MaybeVector2(1, end)
        return MaybeVector2(1, point - me.vector)

    cpdef Vector2 refpoint(SimObject self):
        """The refpoint of self is the point at which forces affect self. This
        is the center of mass of self (if self.isLine(), it is the middle of the
        line.)"""
        if not self.isLine():
            return self.start
        return self.start + self.slope.scale(0.5)
        
    def __str__(self):
        return ("{round_edges: %s,\nstart: %s,\nend: %s,\nslope: %s\nlength: %s,\ndata: %s\n}" %
                (self.round_edges, self.start, self.end, self.slope, self.length, self.data))

    def __repr__(self):
        return "<SimObject: %s>" % (self,)

    def dict(self):
        """Serialize SimObject to a dict containing keys: start, data, round_edges, end, [slope, length]."""
        
        d = {"start": self.start.dict(), "data": self.data.dict(), "round_edges": self.round_edges}
        if self.end is not None:
            d["end"] = self.end.dict()
            d["slope"] = self.slope.dict()
            d["length"] = self.length
        else:
            d["end"] = None
        return d

    def plot_dict(self):
        return self.dict()
        
    def tag(self, name):
        """Tag metadata with a name."""
        
        self.data.tag(name)

    cpdef tick(SimObject self):
        """Remember current force as the previous force, and reset current force."""
        cdef int i, hooks
        hooks = len(self.hooks)
        for i in range(hooks):
            self.hooks[i](self)
        self.data.tick()

    cpdef addForce(SimObject self, Vector2 force):
        """Affect self by a force."""
        self.data.addForce(force)

    cpdef object get_property(SimObject self, str property_name):
        attr = self.data.meta.get(property_name, None)
        if attr is None and self.group is not None:
            return self.group.get_property(property_name)
        return attr

    cpdef enforce_speed_limit(SimObject self, double dt):
        cdef double limit
        try:
            limit = self.get_property("max_speed")
            self.data.velocity.iscale_down(limit)
        except TypeError:
            pass

    cpdef set_rotation(SimObject self, double angle):
        """Set the rotation of a line, do nothing if a point. Set the rotation
        around refpoint to angle."""
        cdef double sina, cosa, halflength, ypos, xpos
        cdef Vector2 ref
        if not self.isLine():
            return
        sina = sin(angle)
        cosa = cos(angle)
        halflength = self.slope.length() / 2
        ypos = sina * halflength
        xpos = cosa * halflength
        ref = self.refpoint()
        self.end = ref + Vector2(xpos, ypos)
        self.start = ref - Vector2(xpos, ypos)
        self._calcSlope()
        
    cpdef rotate(SimObject self, double angle):
        """Rotate around refpoint."""
        self.set_rotation(self.slope.angle() + angle)
            
def dict_to_sim_object(d):
    """Create a SimObject from a dict containing keys: start, end, round_edges, [data]."""
    start = dict_to_vector2(d["start"])
    if d["end"] is None:
        end = None
    else:
        end = dict_to_vector2(d["end"])
    so = SimObject(start, end, round_edges=d.get("round_edges", False))
    if d.has_key("data"):
        data = dict_to_sim_object_data(d["data"])
        so.data = data
    return so
    
def dict_to_vector2(d):
    """Create a Vector2 from a dict containing keys: x, y."""
    return Vector2(d['x'], d['y'])

def dict_to_sim_object_data(d):
    """Create a SimObjectData from a dict containing keys: now, previously, meta."""
    so = SimObjectData()
    if d.has_key("now"):
        so.now = dict_to_vector2(d['now'])
    if d.has_key("previously"):
        so.previously = dict_to_vector2(d['previously'])
    if d.has_key("velocity"):
        so.velocity = dict_to_vector2(d["velocity"])
    so.meta = d['meta']
    return so

cdef class PotentialTerm:
    """A PotentialTerm defines the interface used by ForceLaw to calculate the force that
    SimObjects exert at a certain point."""
    cdef public double c, sigma

    def __init__(self, double c, double sigma):
        self.c = c
        self.sigma = sigma

    cpdef Vector2 forceFrom(self, SimObject obj, Vector2 point):
        """The force that obj exerts at point.

        Calculated as:
        F = (c / (|r| ** sigma)) * (r / |r|)"""

        cdef MaybeVector2 distance = obj.distance(point)
        if not distance.isVector:
            return Vector2(0, 0) # No force
        cdef double length = distance.vector.length()
        cdef double magnitude = self.c / (length ** self.sigma)
        cdef Vector2 direction = distance.vector.normalize()
        return direction.iscale(magnitude)

    def dict(self):
        """Serialize PotentialTerm to a dict containing keys: c, sigma, type."""
        return {"c": self.c, "sigma": self.sigma, "type": "Simple"}

    def __str__(self):
        return "(%s / (|r| ** %s)) * (r / |r|)" % (self.c, self.sigma)

    def __repr__(self):
        return "<PotentialTerm: %s>" % (self,)

cdef class MindistancePotential(PotentialTerm):
    """This potential only has effect within a minimum range, given by
    mindistance."""
    
    cdef public double mindistance
    
    def __init__(self, c, sigma, mindistance):
        PotentialTerm.__init__(self, c, sigma)
        self.mindistance = mindistance

    def dict(self):
        """Serialize MindistancePotential to a dict containing keys: c, sigma, type, mindistance."""

        d = PotentialTerm.dict(self)
        d["type"] = "MindistancePotential"
        d["mindistance"] = self.mindistance
        return d

    cpdef Vector2 forceFrom(MindistancePotential self, SimObject obj, Vector2 point):
        cdef MaybeVector2 distance = obj.distance(point)
        if distance.vector.length() < self.mindistance:
            return PotentialTerm.forceFrom(self, obj, point)
        else:
            return Vector2(0, 0)

cdef class MaxdistancePotential(PotentialTerm):
    """This potential only has effect up to a maximum range, given by
    maxdistance."""

    cdef public double maxdistance

    def __init__(self, c, sigma, maxdistance):
        PotentialTerm.__init__(self, c, sigma)
        self.maxdistance = maxdistance

    def dict(self):
        """Serialize MaxdistancePotential to a dict containing keys: c, sigma, type, maxdistance."""

        d = PotentialTerm.dict(self)
        d["type"] = "MaxdistancePotential"
        d["maxdistance"] = self.maxdistance
        return d

    cpdef Vector2 forceFrom(MaxdistancePotential self, SimObject obj, Vector2 point):
        cdef MaybeVector2 distance = obj.distance(point)
        if not distance.isVector or self.maxdistance <= distance.vector.length():
            return Vector2(0, 0)
        return PotentialTerm.forceFrom(self, obj, point)

cdef class MaxForcePotential(PotentialTerm):
    """This potential will only return forces shorter than a certain length."""

    cdef public double maxforce
    
    def __init__(self, c, sigma, maxforce):
        PotentialTerm.__init__(self, c, sigma)
        self.maxforce = maxforce

    def dict(self):
        d = PotentialTerm.dict(self)
        d["type"] = "MaxForcePotential"
        d["maxforce"] = self.maxforce
        return d

    cpdef Vector2 forceFrom(MaxForcePotential self, SimObject obj, Vector2 point):
        cdef MaybeVector2 distance = obj.distance(point)
        cdef Vector2 force = PotentialTerm.forceFrom(self, obj, point)
        return force.iscale_down(self.maxforce)

cdef class TangentialPotential(PotentialTerm):

    cdef public double angle_factor
    def __init__(self, c, sigma, angle_factor):
        PotentialTerm.__init__(self, c, sigma)
        self.angle_factor = angle_factor

    cpdef Vector2 forceFrom(TangentialPotential self, SimObject obj, Vector2 point):
        cdef Vector2 force = PotentialTerm.forceFrom(self, obj, point)
        cdef Vector2 normal = force.perpendicular()
        normal.iscale(self.angle_factor)
        return force + normal

    def dict(self):
        d = PotentialTerm.dict(self)
        d['type'] = 'TangentialPotential'
        d['angle_factor'] = self.angle_factor
        return d
        
POTENTIAL_TYPES = {
    "Simple": (PotentialTerm, ("c", "sigma")),
    "MindistancePotential": (MindistancePotential, ("c", "sigma", "mindistance")),
    "MaxdistancePotential": (MaxdistancePotential, ("c", "sigma", "maxdistance")),
    "MaxForcePotential": (MaxForcePotential, ("c", "sigma", "maxforce")),
    "TangentialPotential": (TangentialPotential, ("c", "sigma", "angle_factor"))
}

def add_potential_type(name, tp, arg_tuple):
    """Make module aware that name maps to type and is instanciated
    with the dictionary keys contained in arg_tuple."""

    POTENTIAL_TYPES[name] = (tp, arg_tuple)
    
def dict_to_potential(d):
    """Look up known potential types and instanciate from args contained
    in d."""

    tp, args = POTENTIAL_TYPES[d["type"]]
    return tp(*([d[key] for key in args]))

def makeRectangle(startx, starty, xlength, ylength):
    lowerLeft = Vector2(startx, starty)
    upperLeft = Vector2(startx, starty + ylength)
    upperRight = Vector2(startx + xlength, starty + ylength)
    lowerRight = Vector2(startx + xlength, starty)
    return [SimObject(lowerLeft, upperLeft),
            SimObject(lowerLeft, lowerRight),
            SimObject(upperLeft, upperRight),
            SimObject(lowerRight, upperRight)]

def makeSquare(startx, starty, sidelength):
    return makeRectangle(startx, starty, sidelength, sidelength)
    
cdef class ForceLaw:
    """Container for potentials. Potentials must adhere to the interface provided by
    PotentialTerm, that is they must provide the following methods with the signatures:
    - Vector2 forceFrom(self, SimObject, Vector2)
    - dict dict(self)

    Potentials do not need to inherit PotentialTerm, no typechecking is done."""
    cdef public object terms

    def __init__(self, terms):
        self.terms = terms

    cpdef forceFrom(self, SimObject obj, Vector2 point):
        """Sum up the forces from all the terms in this forcelaw."""
        
        cdef Vector2 force = Vector2(0, 0)
        cdef int nterms, i
        nterms = len(self.terms)
        for i in range(nterms):
            force += self.terms[i].forceFrom(obj, point)
        return force

    cpdef affect(self, SimObject source, SimObject target):
        """Register the force from source to target in targets data."""

        cdef Vector2 force = self.forceFrom(source, target.refpoint())
        target.addForce(force)

    def dict(self):
        return {"terms": [term.dict() for term in self.terms]}

    def __str__(self):
        return " + ".join([str(term) for term in self.terms])

    def __repr__(self):
        return "<ForceLaw: %s>" % (self,)

    cpdef affect_from_all(ForceLaw self, list sources, list targets):
        cdef SimObject source, target
        cdef int i, j, imax, jmax
        imax = len(sources)
        jmax = len(targets)
        for i in range(imax):
            source = sources[i]
            for j in range(jmax):
                target = targets[j]
                if source != target:
                    self.affect(source, target)
                    
def dict_to_force_law(d):
    """Create a ForceLaw from a dict containing keys: terms."""
    
    terms = [dict_to_potential(term) for term in d.get("terms", [])]
    return ForceLaw(terms)

cdef class MaxSpeedHook:
    cdef public double speed

    def __init__(self, speed):
        self.speed = speed

    def __call__(self, SimObject obj):
        obj.data.now.iscale_down(self.speed)

cpdef immediate_actuator(SimObject obj, double dt):
    obj.data.velocity = obj.data.now.copy()

cpdef accelerating_actuator(SimObject obj, double dt):
    obj.data.velocity += obj.data.now.scale(dt)

cpdef maximum_distance_pair(list simobjects):
    cdef double longest, current
    cdef SimObject one, two
    cdef int i, j, length
    length = len(simobjects)
    longest = 0
    for i in range(0, length):
        for j in range(i + 1, length):
            current = simobjects[i].distance(simobjects[j].refpoint()).vector.length()
            if current > longest:
                one = simobjects[i]
                two = simobjects[j]
                longest = current
    return one, two

cpdef list objects_near(list SimObjects, Vector2 location, double radius):
    cdef list result = []
    cdef int i
    cdef int length = len(SimObjects)
    cdef double rsquared = radius * radius
    cdef MaybeVector2 distance
    for i in range(length):
        distance = SimObjects[i].distance(location)
        if distance.isVector and distance.vector.length_squared() < rsquared:
            result.append(SimObjects[i])
    return result

cpdef tuple bounding_box(list SimObjects, double radius):
    """Return the bounding box that bounds all SimObjects with radius,
    fully around their refpoints. The result is (lower_left, upper_right)."""
    cdef double ymax, ymin, xmax, xmin
    cdef int i
    cdef int length = len(SimObjects)
    cdef SimObject so
    cdef Vector2 refpoint
    ymax = ymin = SimObjects[0].refpoint().y
    ymax += radius
    ymin -= radius
    xmax = xmin = SimObjects[0].refpoint().x
    xmax += radius
    xmin -= radius
    for i in range(1, length):
        so = SimObjects[i]
        refpoint = so.refpoint()
        if refpoint.x - radius < xmin:
            xmin = refpoint.x - radius
        elif refpoint.x + radius > xmax:
            xmax = refpoint.x + radius
        if refpoint.y - radius < ymin:
            ymin = refpoint.y - radius
        elif refpoint.y + radius > ymax:
            ymax = refpoint.y + radius
    return (Vector2(xmin, ymin), Vector2(xmax, ymax))

cpdef Vector2 bounding_box_side_lengths(list SimObjects, double radius):
    """The side lengths of bounding_box(SimObjects, radius) as a Vector2."""
    cdef Vector2 lower_left, upper_right
    lower_left, upper_right = bounding_box(SimObjects, radius)
    return Vector2(upper_right.x - lower_left.x, upper_right.y - lower_left.y)

cpdef Vector2 center_of_mass(list SimObjects):
    """Locate the center of mass of argument, which may be a list of SimObjects
    or Vector2s."""

    cdef Vector2 total = Vector2(0, 0)
    cdef int i, length
    length = len(SimObjects)
    for i in range(length):
        if isinstance(SimObjects[i], Vector2):
            total += SimObjects[i]
        else:
            total += SimObjects[i].refpoint()
    return total.divide_by(length)