"""Toolkit to generate input data for simulation.py."""
import random
from apf.lib.ccalcs import Vector2, SimObject, ForceLaw, PotentialTerm
from apf.sims.groups import SimObjectGroup, GroupManager
import sys
import apf

def random_vector(xmin, xmax, ymin, ymax):
    return Vector2(random.uniform(xmin, xmax),
                   random.uniform(ymin, ymax))

def distance_constraint(distance):
    def constraint(accepted, candidate):
        for vector in accepted:
            if (vector - candidate).length() < distance:
                return False
        return True
    return constraint

def put_in_simulation_constraint(sim, distance):
    dist_constraint = distance_constraint(distance)
    def constraint(accepted, candidate):
        for group in sim.groupmanager.groups.values():
            if group.objects_near(candidate, distance):
                return False
        return dist_constraint(accepted, candidate) 
    return constraint
    
def n_random_vectors_constrained(xmin, xmax, ymin, ymax, n, constraint):
    vectors = []
    while len(vectors) < n:
        v = random_vector(xmin, xmax, ymin, ymax)
        if constraint(vectors, v):
            vectors.append(v)
    return vectors

def n_random_vectors(xmin, xmax, ymin, ymax, n):
    return [random_vector(xmin, xmax, ymin, ymax) for _ in xrange(n)]
    
class Generator(object):

    def __init__(self, xmin, xmax, ymin, ymax):
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.constraint = None
        self.internal = None
        self.external = None

    def n_objects(self, n):

        if self.constraint is None:
            positions = n_random_vectors(self.xmin, self.xmax, self.ymin,
                                         self.ymax, n)
        else:
            positions = n_random_vectors_constrained(self.xmin, self.xmax,
                                                     self.ymin, self.ymax, n,
                                                     self.constraint)
        return map(SimObject, positions)

    def min_distance(self, dist):
        self.constraint = distance_constraint(dist)

    def group(self, n_objects, tag=None):
        return SimObjectGroup(self.n_objects(n_objects), self.internal,
                              self.external, tag)

    def group_manager(self, n_objects, relations=None):
        group = self.group(n_objects)
        return GroupManager({group.tag: group}, relations)

class BoxGenerator(Generator):
    
    def __init__(self, x, y, size):
        super(BoxGenerator, self).__init__(x, x + size, y, y + size)

    
USAGE = \
"""%s n mindistance

Print simulation.py infile with n objects that are separated by at least mindistance.

The obects are spread uniformly, within a [x, y, side_length] defined in apfrc |
~/.apfrc | apf._SETTINGS in section generate, with key box (see apf_settings.py).
"""

def usage():
    print USAGE % (sys.argv[0], )
    sys.exit()
    
def main():
    if len(sys.argv) != 3:
        usage()
    try:
        n, mindistance = int(sys.argv[1]), float(sys.argv[2])
    except ValueError:
        usage()
    bg = BoxGenerator(*apf._get_option("generate", "box"))
    bg.min_distance(mindistance)
    bg.internal = ForceLaw([PotentialTerm(20, 2), PotentialTerm(-1, 1)])
    print bg.group_manager(n).to_json()

