"""Manage groups of simulation objects."""
import apf.lib.ccalcs as C
import json
import pprint
import apf

class SimObjectGroup(object):
    """A container for SimObjects that can define internal and external forcelaws."""

    TAGNO = 1
    
    def __init__(self, objects, internal=None, external=None, tag=None, data=None):
        self.objects = objects
        for obj in self.objects:
            obj.group = self
        self.internal = internal or C.ForceLaw([])
        self.external = external or C.ForceLaw([])
        if tag is None:
            self.tag = str(SimObjectGroup.TAGNO)
            SimObjectGroup.TAGNO += 1
        else:
            self.tag = tag
        self.hooks = []
        if data is None:
            self.data = {}
        else:
            self.data = data
        
    def add(self, obj):
        obj.group = self
        self.objects.append(obj)

    def run_hooks(self):
        for hook in self.hooks:
            hook(self)
        
    def tick(self, runhooks=True):
        if runhooks:
            self.run_hooks()
        for obj in self.objects:
            obj.tick()

    def calcInternalForces(self):
        self.internal.affect_from_all(self.objects, self.objects)

    def moveAllObjects(self, dt):
        for obj in self.objects:
            obj.actuate(dt)
            obj.enforce_speed_limit(dt)
            obj.move(obj.data.velocity.scale(dt))

    def affectOther(self, otherGroup):
        self.external.affect_from_all(self.objects, otherGroup.objects)

    def dict(self):
        return {"objects": [obj.dict() for obj in self.objects],
                "internal": self.internal.dict(),
                "external": self.external.dict(),
                "tag": self.tag,
                "data": self.data}

    def plot_dict(self):
        return {"objects": [obj.plot_dict() for obj in self.objects]}

    @staticmethod
    def from_dict(d):
        return SimObjectGroup([C.dict_to_sim_object(obj) for obj in d["objects"]],
                              C.dict_to_force_law(d["internal"]),
                              C.dict_to_force_law(d["external"]),
                              d["tag"],
                              data=d.get("data", {})
                              )

    @staticmethod
    def from_json(js):
        return SimObjectGroup.from_dict(json.loads(js))
        
    def to_json(self):
        return json.dumps(self.dict(), indent=apf._get_option("all", "indent"))

    def unitByTag(self, value, tag="name"):
        for obj in self.objects:
            if obj.data.meta[tag] == value:
                return obj

    def add_hook(self, hook):
        self.hooks.append(hook)

    def get_property(self, name):
        return self.data.get(name, None)

    def internal_field_strength_in(self, loc):
        return self.field_strength_in(loc, self.internal)

    def external_field_strength_in(self, loc):
        return self.field_strength_in(loc, self.external)

    def field_strength_in(self, loc, potential):
        force = C.Vector2(0, 0)
        for obj in self.objects:
            force += potential.forceFrom(obj, loc)
        return force

    def objects_near(self, loc, radius):
        return C.objects_near(self.objects, loc, radius)

    def kill_unit(self, simobj):
        self.objects.remove(simobj)
        
class GroupManager(object):
    """A group-manager stores groups and their relations (Which groups that exert forces on others)."""

    def __init__(self, groups, relations=None):
        self.relations = relations or set([])
        self.groups = groups
        self.hooks = []

    def add_hook(self, hook):
        self.hooks.append(hook)
        
    def dict(self):
        return {"groups": dict([(g.tag, g.dict()) for g in self.groups.values()]),
                "relations": list(self.relations)}

    def plot_dict(self):
        return {"groups": dict([(g.tag, g.plot_dict()) for g in self.groups.values()])}
        
    def add_relation(self, source, target):
        self.relations.add((source.tag, target.tag))

    @staticmethod
    def from_dict(d):
        groupTags = d["groups"].keys()
        groups = dict(d["groups"])
        for groupTag in groups:
            groups[groupTag] = SimObjectGroup.from_dict(groups[groupTag])
        for relation in d["relations"]:
            assert relation[0] in groups and relation[1] in groups
        return GroupManager(groups, set([tuple(rel) for rel in d["relations"]]))

    def to_json(self):
        return json.dumps(self.dict(), indent=apf._get_option("all", "indent"))

    @staticmethod
    def from_json(js):
        return GroupManager.from_dict(json.loads(js))

    def run_hooks(self):
        for hook in self.hooks:
            hook(self)
            
    def tick(self, runhooks=True):
        if runhooks:
            self.run_hooks()
        for group in self.groups.values():
            group.tick()

    def move(self, dt):
        for group in self.groups.values():
            group.moveAllObjects(dt)

    def calcForces(self):
        for group in self.groups.values():
            group.calcInternalForces()
        for source, target in self.relations:
            self.groups[source].affectOther(self.groups[target])

    def dump(self, fileObject):

        fileObject.write(self.to_json())

    def dump_pretty(self, fileObject):

        pprint.pprint(self.to_json(), fileObject)

    def unit_and_group_by_tag(self, value, tag="name"):
        for group in self.groups.values():
            unit = group.unitByTag(value, tag)
            if unit:
                return unit, group

    def unit_by_tag(self, value, tag="name"):
        return self.unit_and_group_by_tag(value, tag)[0]
        
    def force_in(self, loc, grouptag):
        force = self.groups[grouptag].internal_field_strength_in(loc)
        for relation in self.relations:
            source, dest = relation
            if dest == grouptag:
                force += self.groups[source].external_field_strength_in(loc)
        return force

    def force_grid(self, grid, grouptag):
        for pos in grid:
            yield pos, self.force_in(pos, grouptag)
            