"""
Setting different scoring and stopping criteria for simulations that
need to terminate without dumping output.

Also provides operations for mutation on strands/genomes of simulations
such that they can be used for inspyreds multitude of optimization
techniques with only defining goals and constraints.
"""

from apf.sims import simulation
from apf import custom
import copy
import inspyred
import random
import multiprocessing
import Queue

class ScoreCounter(object):

    def __init__(self, score_f, finalize_f):
        self.score_f = score_f
        self.finalize_f = finalize_f
        self._score = 0
        
    def count(self, sim):
        self._score += self.score_f(sim)
        self.score(sim)

    def score(self, sim):
        sim.score = self.finalize_f(self._score, sim)
        return sim.score

class TerminatingSimulation(object):

    def __init__(self, sim, scorecounter, termination_condition):

        self.sim = sim
        self.sim.disable_dump = True
        self.scorecounter = scorecounter
        self.termination_condition = termination_condition

    def done(self):

        return self.termination_condition(self)

    def score(self):

        return self.scorecounter.score
        
    def tick(self):

        self.sim.tick()
        self.scorecounter.count(self.sim)

    def run(self):

        while not self.done():
            self.tick()

def force_law_getter(group_name, get='internal'):
    """Return a callable that automatically extracts a force law from a simulation."""
    def _get(sim):
        group = sim.groupmanager.groups[group_name]
        return getattr(group, get, None)
    return _get

TASK_PARAMS = multiprocessing.Queue()
RESULT_PARAMS = multiprocessing.Queue()
PROCESSES = []

def mp_evaluate(inq, outq, evaluate_single):
    while True:
        queue_item = inq.get()
        if queue_item == 'stop':
            break
        index, params = queue_item
        performance = evaluate_single(params)
        outq.put((index, performance))

def sc_evaluate_generation(candidates, args):
    evaluator = args["evaluate_single"]
    return [evaluator(candidate) for candidate in candidates]
   
def setup_mp(inq, outq, evaluator, numcpus=3):
    for i in range(numcpus):
        PROCESSES.append(multiprocessing.Process(target=mp_evaluate, args=(inq, outq, evaluator)))
        PROCESSES[-1].start()

def stop_mp():
    for p in PROCESSES:
        TASK_PARAMS.put("stop")

def mp_evaluate_generation(candidates, args):
    evaluator = args['evaluate_single']
    for index, candidate in enumerate(candidates):
        TASK_PARAMS.put((index, candidate))
    results = []
    while len(results) < len(candidates):
        try:
            index, params = TASK_PARAMS.get(True, 0.5)
            performance = evaluator(params)
            results.append((index, performance))
        except Queue.Empty:
            pass
        index, performance = RESULT_PARAMS.get()
        results.append((index, performance))
    results.sort(key=lambda item: item[0])
    return [score for index, score in results]

class InspyredInterface(object):

    def __init__(self, simulation_creator):

        self.lower_bound = []
        self.upper_bound = []
        self.creator = simulation_creator
        self.fl_getters = []
        self.params = {}

    def add_param(self, param, value):
        self.params[param] = value
        
    def add_train_force_law(self, c_bounds, sigma_bounds, group, get='internal'):
        """Add a force law to train.

        c_bounds is a list of (low, high) bounds for the c values in the potential terms for force law.
        sigma_bounds is a list of (low, high) bounds for the sigma values in the potential terms for force law.
        group is the tag of the group that has this force law
        get is either internal for an internal force law or external for an external force law belonging to group."""
        self._add_bounds(copy.copy(c_bounds), copy.copy(sigma_bounds))
        self.fl_getters.append(force_law_getter(group, get))
        
    def _add_bounds(self, cs, sigmas):
        
        while cs and sigmas:
            lowc, hic = cs.pop(0)
            self.lower_bound.append(lowc)
            self.upper_bound.append(hic)
            lowsigma, hisigma = sigmas.pop(0)
            self.lower_bound.append(lowsigma)
            self.upper_bound.append(hisigma)

    def get_bounder(self):

        return inspyred.ec.Bounder(self.lower_bound, self.upper_bound)
        
    def apply(self, reals, sim):

        index = 0
        for getter in self.fl_getters:
            for term in getter(sim).terms:
                term.c, term.sigma = reals[index:index + 2]
                index += 2

    def evaluate_sim(self, candidate, args):

        sim = self.creator()
        self.apply(candidate, sim)
        sim.disable_dump = args.get('disable_dump', True)
        sim.run()
        score = sim.score 
        sim_printing_hook = args.get('sim_printing_hook')
        if sim_printing_hook is not None:
            sim_printing_hook(candidate, score, sim)
        return score

    def evaluator(self, candidates, args):
        return [self.evaluate_sim(candidate, args) for candidate in candidates]

    def generator(self, random, args):
        genome = []
        for (low, high) in zip(self.lower_bound, self.upper_bound):
            low, high = float(low), float(high)
            mean = (low + high) / 2
            stddev = abs(low - high) / 5
            genome.append(random.gauss(mean, stddev))
        if args.get("print_generated", False):
            print genome
        return genome

    def make_kwargs(self):

        needed = {"generator": self.generator,
                  "bounder": self.get_bounder(),
                  "evaluator": self.evaluator}
        needed.update(self.params)
        return needed

    def copy(self):
        return copy.deepcopy(self)

    
    def apply_and_run(self, ec, evaluator, mp=True, startmp=True, numcpus=3, stopmp=True):
        """Take an evolutionary computation and a function that evaluates a single candidate. If mp is True,
        make a process-pool of numcpus that share the load of evaluating evaluator."""
        kwargs = self.make_kwargs()
        for key in ["terminator", "variator", "observer", "archiver"]:
            if key in kwargs:
                setattr(ec, key, kwargs[key])
        if mp:
            if startmp:
                setup_mp(TASK_PARAMS, RESULT_PARAMS, evaluator, numcpus)
            evaluate_many = mp_evaluate_generation
        else:
            evaluate_many = sc_evaluate_generation
        kwargs['evaluate_single'] = evaluator
        kwargs['evaluator'] = evaluate_many
        try:
            pop = ec.evolve(**kwargs)
        finally:
            if mp and stopmp:
                stop_mp()
        return pop
            
def best_from(inspyred_classes, params_list, rand_seed):
    cls_names = []
    results = []
    for params, cls in zip(params_list, inspyred_classes):
        if isinstance(cls, str):
            cls_names.append(cls)
            cls = getattr(inspyred, cls)
        else:
            cls_names.append(cls.__name__)
        r = random.Random()
        r.seed(rand_seed)
        instance = cls(r)
        kwargs = params.make_kwargs()
        kwargs["statistics_file"] = open("%s%s.csv" % (cls_names[-1], rand_seed), "w")
        kwargs["individuals_file"] = open("%s%s-individuals.csv" % (cls_names[-1], rand_seed), "w")
        for key in ["terminator", "variator", "observer", "archiver"]:
            if key in kwargs:
                print "set", key, "=", kwargs[key]
                setattr(instance, key, kwargs[key])
        population = instance.evolve(**kwargs)
        results.append((max(population).candidate, max(population).fitness))
    return cls_names, results
                       
def apply_genome(genome, simulation, force_law_getters):
    index = 0
    for getter in force_law_getters:
        for term in getter(simulation).terms:
            term.c, term.sigma = genome[index:index + 2]
            index += 2