"""
Plot data from a simulation.
"""
import json
import matplotlib
import sys
import time
from multiprocessing import Process
import apf
import math
import numpy as np

for backend in apf._get_option("plot", "backends"):
    try:
        matplotlib.use(backend)
        break
    except:
        sys.stderr.write("Warning: %s not available as backend\n" % (backend,))

import matplotlib.pyplot as plt
import matplotlib.patches as patches

def want_frame(frame):
    return frame["time"] % apf._get_option("plot", "plot_frames") == 0
    
def get_frames(infile):
    if isinstance(infile, file):
        content = infile.read()
    else:
        with open(infile, "rb") as fp:
            content = fp.read()
    frames = json.loads(content)
    header = frames[0]
    header["groups"] = header["groupmanager"]["groups"]
    header["time"] = 0
    header["dt"] = 0.0
    wanted = [frame for frame in frames if want_frame(frame)]
    for index, frame in enumerate(wanted):
        frame["time"] = index
    return wanted

def get_objects_by_look(framedict):
    out = {}
    for grouptag in framedict["groups"]:
        objects = framedict["groups"][grouptag]["objects"]
        for obj in objects:
            meta = obj["data"]["meta"]
            if not meta.get("draw", True):
                continue
            try:
                out[meta["plt"]] = out.get(meta["plt"], []) + [obj]
            except KeyError:
                # Default to blue circle
                default = apf._get_option("plot", "default_plt")
                out[default] = out.get(default, []) + [obj]
    return out

def plot_framedict(framedict):
    objs = get_objects_by_look(framedict)
    for look in objs:
        plot_look(look, objs[look])

def _get_position(objdict):
    pos = objdict["start"]
    return pos["x"], pos["y"]

def _get_force(objdict):
    force = objdict["data"]["now"]
    return force["x"], force["y"]

def _get_velocity(objdict):
    velocity = objdict['data']['velocity']
    return velocity['x'], velocity['y']

def draw_circle(center, radius, color):
    circle = patches.Circle(center, fill=True, radius=radius, color=color)
    plt.axes().add_patch(circle)
    
def plot_look(look, objs):
    lines = [obj for obj in objs if obj["end"] is not None]
    points = [obj for obj in objs if obj["end"] is None]
    if apf._get_option("plot", "quivers"):
        for point in points:
            x, y = _get_position(point)
            fx, fy = _get_velocity(point)
            plt.quiver([x], [y], [fx], [fy], color=look[0], scale=apf._get_option("plot", "quiver_scale", 1))
    if apf._get_option("plot", "aquivers"):
        for point in points:
            x, y = _get_position(point)
            fx, fy = _get_force(point)
            plt.quiver([x], [y], [fx], [fy], color=look[0], scale=apf._get_option("plot", "aquiver_scale", 1))
    for obj in points:
        radius = obj["data"]["meta"].get("radius", 4)
        x, y = obj["start"]["x"], obj["start"]["y"]
        color = look[0]
        draw_circle((x, y), radius, color)
    for line in lines:
        x1, x2 = line["start"]["x"], line["end"]["x"]
        y1, y2 = line["start"]["y"], line["end"]["y"]
        plt.plot([x1, x2], [y1, y2], look[0] + "-")

def plot_frame(name_prefix, frame):
    plt.axis(apf._get_option("plot", "axis"), 'scaled')
    plot_framedict(frame)
    plt.savefig("%s%04d" % (name_prefix, frame["time"]))
    plt.clf()

def plot_frames(name_prefix, frames):
    for frame in frames:
        print "Plotting frame: %d, dt: %.3f" % (frame["time"], frame["dt"])
        plot_frame(name_prefix, frame)

def plot_file(name_prefix, infile):
    frames = get_frames(infile)
    plot_frames(name_prefix, frames)

def partition(sliceable, parts):
    return [sliceable[i::parts] for i in xrange(parts)]

USAGE = """%s <infile> <prefix> [nthreads] | <prefix>

Generate png-plots from a datadump made by simulation.py. Files are stored in
prefix%%4d.png in current working directory. Run nthreads plotting in parallell.

Take data from stdin if there is no infile given on command line.
"""

def usage(progname):
    sys.stderr.write(USAGE % (progname,))

def run(prefix, framegroups, processes):
    ps = []
    for framegroup in framegroups[:-1]:
        p = Process(target=plot_frames, args=(prefix, framegroup))
        ps.append(p)
        p.start()
    plot_frames(prefix, framegroups[-1])
    for p in ps:
        p.join()

def get_framegroups(infile, processes):
    return partition(get_frames(infile), processes)

def plot_unit_hook(colorspec, shape="o", scale=100):
    def hook(obj):
        x, y = obj.start.x, obj.start.y
        if obj.isLine():
            plt.plot([x, obj.end.x], [y, obj.end.y], colorspec + '-')
        else:
            plt.plot([x], [y], colorspec + shape)
        fx, fy = obj.data.now.x, obj.data.now.y
        fx *= scale
        fy *= scale
        plt.quiver([x], [y], [fx], [fy], angles='xy', scale=scale, color='r')
    return hook

def figure_saver_hook(prefix):
    def hook(sim):
        plt.savefig("%s%04d.png" % (prefix, sim.time))
        plt.clf()
    return hook
    
def attach_plotter(simulation, prefix, scale=100):
    for group in simulation.groupmanager.groups.values():
        for obj in group.objects:
            obj.add_hook(plot_unit_hook(obj.data.meta.get("plt", "bo"), "", scale=scale))
    simulation.add_hook(figure_saver_hook(prefix))

def vfield_plotter_hook(grouptag, grid, color, scale=1000):
    def hook(simulation):
        fgrid = simulation.groupmanager.force_grid(grid, grouptag)
        for pos, force in fgrid:
            x, y = pos.x, pos.y
            fx, fy = force.x, force.y
            plt.quiver([x], [y], [fx], [fy], angles='xy', scale=scale, color=color)
    return hook
    
def _almost(x, y, eps=1E-7):
    """Float comparison."""
    return abs(x - y) <= eps

    
def main():
    if not len(sys.argv) in (2, 3, 4):
        usage(sys.argv[0])
        sys.exit()
    infile = sys.argv[1]
    if len(sys.argv) == 2:
        infile = sys.stdin
        prefix = sys.argv[1]
    else:
        prefix = sys.argv[2]
    try:
        processes = int(sys.argv[3])
    except IndexError:
        processes = apf._get_option("plot", "processes")
    framegroups = partition(get_frames(infile), processes)
    run(prefix, framegroups, processes)