"""Run a simulation and dump output to a file."""
import json
import groups
import sys
import apf.lib.ccalcs as C
from apf import utils
import apf

class SimulationParameters(object):
    def __init__(self, infile, timestep, outfile, steps):
        
        self.infile = infile
        if infile == "sys.stdin":
            self.infile = sys.stdin
        self.timestep = timestep
        self.outfile = outfile
        if outfile == "sys.stdout":
            self.outfile = sys.stdout
        self.steps = steps
        
    def dict(self):
        d = {"infile": self.infile, "timestep": self.timestep,
             "outfile": self.outfile, "steps": self.steps}
        if self.infile is sys.stdin:
            d["infile"] = "sys.stdin"
        if self.outfile is sys.stdout:
            d["outfile"] = "sys.stdout"
        return d

    def to_json(self):
        return json.dumps(self.dict(), indent=apf._get_option("all", "indent"))

    @staticmethod
    def from_dict(d):
        return SimulationParameters(d["infile"], d["timestep"], d["outfile"], d["teps"])

    @staticmethod
    def from_json(js):
        return SimulationParameters.from_dict(json.loads(js))

    @staticmethod
    def from_config():
        inf = apf._get_option("fullsim", "in")
        out = apf._get_option("fullsim", "output")
        steps = apf._get_option("fullsim", "steps")
        dt = apf._get_option("fullsim", "dt")
        return SimulationParameters(inf, dt, out, steps)
        
class Simulation(object):
    
    def __init__(self, params):
        self.params = params
        self.disable_dump = False
        infile = None
        try:
            if not isinstance(params.infile, file):
                try:
                    infile = open(params.infile, "rb")
                except IOError:
                    infile = utils.get_apf_data_file(params.infile)
            else:
                infile = self.params.infile
            data = infile.read()
        finally:
            if infile is not None:
                infile.close()
            else:
                raise IOError("No acceptable input file found for project %s not readable." % (params.infile,))
        self.groupmanager = groups.GroupManager.from_json(data)
        self.log = []
        self.time = 0
        self.dt = 0.0
        if not isinstance(self.params.outfile, file):
            self.outfile = open(params.outfile, "wb")
        else:
            self.outfile = params.outfile
        self.hooks = []
        self._done = False

    def add_hook(self, hookfunc):
        self.hooks.append(hookfunc)
            
    def dict(self):
        return {"params": self.params.dict(),
                "groupmanager": self.groupmanager.dict()}

    def to_json(self):
        return json.dumps(self.dict(), indent=apf._get_option("all", "indent"))

    @staticmethod
    def from_dict(d):
        return Simulation(SimulationParameters.from_dict(d["params"]),
                          groups.GroupManager.from_dict(d["groupmanager"]))

    @staticmethod
    def from_json(js):
        return Simulation.from_dict(json.loads(js))

    def tick(self):
        for hook in self.hooks:
            hook(self)
        self.time += 1
        self.dt += self.params.timestep
        self.groupmanager.calcForces()
        self.groupmanager.move(self.params.timestep)
        if self.time % apf._get_option("simulation", "keep_frames") == 0:
            self.dump()
        self.groupmanager.tick()

    def dump(self):
        if self.disable_dump:
            return
        groups = self.groupmanager.plot_dict()
        dict = {"time": self.time, "dt": self.dt}
        dict.update(groups)
        self.log.append(dict)

    def dump_header(self):
        groups = self.groupmanager.dict()
        self.log.append({"groupmanager": groups, "params": self.params.dict()})
        
    def run(self):

        self.log = []
        self.dump_header()
        while not self.done():
            self.tick()
        if not self.disable_dump:
            self.outfile.write(json.dumps(self.log, indent=apf._get_option("all", "indent")))
        self.outfile.close()

    def done(self):

        return self.time >= self.params.steps or self._done

    def finish(self):
        
        self._done = True
        
USAGE_STR = (
"""
Usage: %s maketemplate <file> | <infile> <timedelta> <outfile> <steps> | <dt> <steps> | list

    In first invocation, make a template for infiles in file. This writes a
    JSON file in the destination. This JSON file shows the data format that is
    expected for input files to program.

    In the second invocation, run a potential field simulation with data from
    infile. Each step of the simulation has timedelta length. The simulation
    dumps output for each step to outfile. Steps decides the amount of loops
    the simulation runs.

    In the third invocation, read indata from sys.stdin and print outdata to sys.stdout.

    In the fourth invocation, list available data files from apf package.
""")

def usage(progname):
    sys.stderr.write(USAGE_STR % (progname,))
    sys.exit()

def make_template(destfile):
    with open(destfile, "wb") as outf:
        obj = C.SimObject(C.Vector2(1, 2), C.Vector2(2, 2), True)
        obj.tag("object name here, if any")
        obj.data.meta["color"] = "other keys and values are possible"
        obj2 = C.SimObject(C.Vector2(2, 4))
        obj2.tag("this object is in another group")
        pt = C.PotentialTerm(30.0, 1)
        maxpt = C.MaxdistancePotential(20.0, 1, 100)
        minpt = C.MindistancePotential(-15.0, 2, 40)
        internal = C.ForceLaw([pt, maxpt])
        external = C.ForceLaw([pt, minpt])
        g1 = groups.SimObjectGroup([obj], internal, None,
        "grouptags will be autogenerated if not present")
        g2 = groups.SimObjectGroup([obj2], None, external,
        "GroupManagers have relations of the form (g1.tag, g2.tag) -> group 1 wil exert forces on group 2.")
        gm = groups.GroupManager({g1.tag: g1, g2.tag: g2}, ((g1.tag, g2.tag),))
        outf.write(gm.to_json())
    sys.exit()

def simulate_as_pipe():
    dt = float(sys.argv[1])
    steps = int(sys.argv[2])
    params = SimulationParameters(sys.stdin, dt, sys.stdout, steps)
    Sim = Simulation(params)
    Sim.run()

def simulate_from_infile():
    infile = sys.argv[1]
    dt = float(sys.argv[2])
    outfile = sys.argv[3]
    steps = int(sys.argv[4])
    params = SimulationParameters(infile, dt, outfile, steps)
    Sim = Simulation(params)
    Sim.run()
    
def main():
    if len(sys.argv) == 3 and sys.argv[1] == "maketemplate":
        make_template(sys.argv[2])
    elif len(sys.argv) == 2 and sys.argv[1] == "list":
        print "\n".join(utils.get_apf_data_list(".json"))
    elif len(sys.argv) == 3:
        simulate_as_pipe()
    elif len(sys.argv) < 5:
        usage(sys.argv[0])
    else:
        simulate_from_infile()
