"""
Visualize a simulation using pygame.
"""

import pygame
from pygame.locals import *
import apf
from apf.sims import simulation, plot
from apf.lib.ccalcs import Vector2
import sys

pygame.init()

RED = (255, 0, 0)
GREEN = (0, 255, 0)
BLUE = (0, 0, 255)
YELLOW = (255, 255, 0)
GOLD = (255, 215, 0)
BLACK = (0, 0, 0)
WHITE = (255, 255, 255)

_COLORMAP = {
    "b": BLUE,
    "r": RED,
    "g": GREEN,
    "y": YELLOW
}

def _pygame_color(simobject):
    meta = simobject.data.meta
    return _COLORMAP.get(meta.get("plt", "b")[0], GOLD)

def _unwrap(vector2):
    return (int(vector2.x), int(vector2.y))
    
class PixelMapper(object):
    """Map a simulation location to a screen location."""
    
    def __init__(self, sim_space, pygame_space):
        """sim_space is a Vector2 with the x, y size of the simulation space.
        pygame_space is a Vector2 with the x, y size of the screen space."""
        
        self.sim_space = sim_space
        self.pygame_space = pygame_space
        self.scale = Vector2(pygame_space.x / sim_space.x, pygame_space.y / sim_space.y)
        
    def to_screen(self, pos):
        """Figure out where pos is on screen."""

        # Origin in simulation is mid of screen but in pygame, it's topleft
        pos = Vector2(pos.x / self.sim_space.x, pos.y / self.sim_space.y) + Vector2(0.5, 0.5)
        pos.y = 1 - pos.y
        return pos * self.pygame_space # Y increases towards bottom of screen

    def scale_to_pygame(self, size):
        """Figure out how large size needs to be to be correct in pygame."""
        return self.scale * size
        
class PygameSimulation(object):

    def __init__(self, sim, screen_size, sim_size):
        sim.disable_dump = True
        self.sim = sim
        self.mapper = None
        self.surf = None
        self.clock = None
        self._setup(screen_size, sim_size)
        self.linewidth = 4
        self.circle_radius = 8
        self.framerate = 40
        
    def _setup(self, screen_size, sim_size):
        self.mapper = PixelMapper(Vector2(*sim_size), Vector2(*screen_size))
        pygame.display.init()
        self.surf = pygame.display.set_mode(screen_size)
        self.clock = pygame.time.Clock()
        
    def start_location(self, sim_object):
        return self.mapper.to_screen(sim_object.start)

    def draw_object(self, sim_object):
        if not sim_object.data.meta.get("draw", True):
            return
        if sim_object.isLine():
            self.draw_line(sim_object)
        else:
            self.draw_circle(sim_object)

    def draw_line(self, sim_object):

        start = _unwrap(self.start_location(sim_object))
        end = _unwrap(self.mapper.to_screen(sim_object.end))
        pygame.draw.line(self.surf, _pygame_color(sim_object), start, end, self.linewidth)

    def draw_circle(self, sim_object):

        center = _unwrap(self.start_location(sim_object))
        pygame.draw.circle(self.surf, _pygame_color(sim_object), center, self.circle_radius)

    def draw_all_objects(self):

        for group in self.sim.groupmanager.groups.values():
            for sim_object in group.objects:
                self.draw_object(sim_object)

    def tick(self):

        self.sim.tick()
        self.surf.fill(WHITE)
        self.draw_all_objects()
        
    def run(self):

        while not self.sim.done():
            self.tick()
            self.do_events()
        self.do_events(hang=True)

    def do_events(self, hang=False):

        poll = True
        while poll:
            poll = hang            
            for event in pygame.event.get():
                if event.type == QUIT:
                    pygame.quit()
                    sys.exit()
            pygame.display.update()
            self.clock.tick(self.framerate)
